﻿using System;
using SWConfigDataClientLib;
using SWConfigDataClientLib.Data;
using SWConfigDataClientLib.Proxies.Admin;
using SWConfigDataClientLib.Proxies.PhoneClient;
using SWConfigDataClientLib.Proxies.Users;
using SWConfigDataClientLib.Proxies.Files;
using SWConfigDataClientLib.WSPhoneClientFacade;
using SWConfigDataSharedLib.Tracing;
using SWConfigDataClientLib.Proxies.UserPhoneBook;

namespace IpPbx.Samples.Shared
{
    /// <summary>
    /// Contains all methods which uses the CDSClient
    /// </summary>
    public class CDSLib : IDisposable
    {
        #region Tracing

        /// <summary>
        /// Init user defined trace module named "CDSlib" 
        /// See App.config or Web.config for trace module configuration
        /// </summary>
        private static STraceInfo log = new STraceInfo(
            new System.Diagnostics.TraceSwitch("CDSLib", "CDSlib", "2"), typeof(CDSLib));

        #endregion

        #region Fields

        /// <summary>
        /// The CDS Client LibManager is the central class of the CDS Client API.
        /// It contains methods to address the server and to authenticate users.
        /// You use LibManager to create instances of classes for retrieval and
        /// manipulation of IpPbx configuration items. The LibManager should be the
        /// first class to instantiate. It checks the configuration and initiates
        /// the tracing of the client application. 
        /// </summary>
        private LibManager libManager;

        /// <summary>
        /// Web Service proxy for IpPbx clients. It contains specialized views
        /// and methods to be used from client application. A PhoneClientFacade
        /// instance is implicitly associated to the IpPbx user specified in
        /// the lib manager.
        /// </summary>
        private PhoneClientFacade phoneClientFacade;

        /// <summary>
        /// Web service proxy for administration clients. It contains specialized
        /// views analogue to the views used in the IpPbx Administration.
        /// If your application does any administrative tasks, look at this object first.
        /// </summary>
        private AdminFacade adminFacade;

        /// <summary>
        /// Web service proxy for managing all files stored in IpPbx's database.
        /// </summary>
        private FilesFacade filesFacade;

        #endregion

        #region Properties

        /// <summary>
        /// Returns the name of the current user. A username is
        /// required for the UsernamePassword authentication mode.
        /// </summary>
        public string UserName
        {
            get
            {
                if (this.libManager == null ||
                    string.IsNullOrWhiteSpace(this.libManager.Username))
                {
                    return null;
                }

                return this.libManager.Username;
            }
        }

        /// <summary>
        /// Returns the server address of the web service.        
        /// </summary>
        public string ServerName
        {
            get
            {
                if (this.libManager == null ||
                    string.IsNullOrWhiteSpace(this.libManager.WSBaseUrl))
                {
                    return null;
                }

                return this.libManager.WSBaseUrl;
            }            
        }

        /// <summary>
        /// Returns the current AuthenticationMode
        /// </summary>  
        public SProxyObject.AuthenticationMode AuthenticationMode
        {
            get
            {
                if (this.libManager == null)
                {
                    return SProxyObject.AuthenticationMode.NoAuthentification;
                }

                return  this.libManager.AuthenticationMode;
            } 
        }

        #endregion

        //------------------------------------------------------------------------------------

        #region CTors

        /// <summary>
        /// Creates a new CDS object (Windows Authentication)
        /// Assumes that the SwyxServer is running on the local machine (localhost)
        /// </summary>        
        public CDSLib()
            : this(String.Empty, String.Empty, String.Empty)
        {
            // Nothing else to do
        }

        /// <summary>
        /// Creates a new CDS object (Windows Authentication)
        /// </summary>     
        /// <param name="serverName">SwyxServer Address (FQDN or IP)</param>
        public CDSLib(string serverName)
            : this(String.Empty, String.Empty, serverName)
        {
            // Nothing else to do
        }

        /// <summary>
        /// Creates a new CDS object (UsernamePassword Authentication)
        /// Assumes that the SwyxServer is running on the local machine (localhost)
        /// </summary>
        /// <param name="userName">Username (SwyxIt Login)</param>
        /// <param name="password">Password (SwyxIt Login)</param>        
        public CDSLib(string userName, string password)
            : this(userName, password, String.Empty)
        {
            // Nothing else to do
        }

        /// <summary>
        /// Creates a new CDS object (UsernamePassword Authentication)
        /// </summary>
        /// <param name="userName">Username (SwyxIt Login)</param>
        /// <param name="password">Password (SwyxIt Login)</param>
        /// <param name="serverName">SwyxServer Address (FQDN or IP)</param>
        public CDSLib(string userName, string password, string serverName)
        {
            this.Init(userName, password, serverName);
        }

        #endregion

        #region Initialisation

        /// <summary>
        /// Initialisation
        /// </summary>
        /// <param name="userName">Username (SwyxIt Login)</param>
        /// <param name="password">Password (SwyxIt Login)</param>
        /// <param name="serverAddress">SwyxServer Address (FQDN or IP)</param>
        private void Init(string userName, string password, string serverAddress)
        {
            log.TraceInfo("Init", "()");

            // Check if server address is empty
            if (String.IsNullOrWhiteSpace(serverAddress))
            {
                serverAddress = "127.0.0.1";
            }

            log.TraceInfo("Init", "ServerAddress: {0}", serverAddress);

            // Set default authentication mode to trusted
            SProxyObject.AuthenticationMode authenticationMode = SProxyObject.AuthenticationMode.Trusted;

            // If username and password are available the authentication
            // mode must be changed to UsernamePassword
            if (!string.IsNullOrWhiteSpace(userName) &&
                !string.IsNullOrWhiteSpace(password))
            {
                authenticationMode = SProxyObject.AuthenticationMode.UsernamePassword;
            }

            log.TraceInfo("Init", "AuthenticationMode: {0}", authenticationMode);

            if (authenticationMode == SProxyObject.AuthenticationMode.UsernamePassword)
            {
                log.TraceInfo("Init", "Username: {0}", userName);
            }

            // Create and configure new LibManager object
            this.libManager = new LibManager()
            {
                BaseProtocol = SProxyObject.BaseProtocol.TCP,
                Username = userName,
                Password = password,
                WSBaseUrl = serverAddress,
                AuthenticationMode = authenticationMode
            };

            log.TraceInfo("Init", "CDSLib initialisation completed");
        }

        #endregion

        #region IDisposable Members

        /// <summary>
        /// Performs application-defined tasks associated with freeing,
        /// releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing,
        /// releasing, or resetting unmanaged resources.
        /// </summary>
        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                // Cleanup
                if (this.adminFacade != null)
                {
                    this.adminFacade.FreeForReuse();
                    this.adminFacade = null;
                }

                if (this.filesFacade != null)
                {
                    this.filesFacade.FreeForReuse();
                    this.filesFacade = null;
                }

                if (this.phoneClientFacade != null)
                {
                    this.phoneClientFacade.FreeForReuse();
                    this.phoneClientFacade = null;
                }

                if (this.libManager != null)
                {
                    this.libManager.FreeForReuse();
                    this.libManager = null;
                }
            }
        }

        #endregion

        #region Test Connection

        /// <summary>
        /// Check if a CDS connection can be established
        /// The use of this method is NOT mandatory.
        /// </summary>
        public bool TestConnection()
        {
            log.TraceInfo("TestConnection", "()");

            return this.HasRole(BuildInRoles.UserSelf);
        }

        #endregion

        //------------------------------------------------------------------------------------
        // SAMPLE METHODS
        //------------------------------------------------------------------------------------
        
        #region Files (Get Skin List & Upload Skin)

        /// <summary>
        /// Return the path to the file in the IpPbx file cache (on local harddisk)
        /// </summary>
        public string GetFile(FileListEntry fileEntry)
        {
            // Ensure that a FilesFacade object is available
            if (this.filesFacade == null)
            {
                this.filesFacade = this.libManager.GetFilesFacade();
            }

            return this.filesFacade.GetIpPbxFile(fileEntry);
        }

        //------------------------------------------------------------------------------------

        /// <summary>
        /// Adds a skin file to the database into the global scope
        /// The skin will be available for every ippbx user
        /// </summary>
        /// <param name="filename">Name of the file in the database</param>
        /// <param name="filePath">Source path from where file should be uploaded</param>
        public void AddSkin(string filename, string filePath)
        {
            // Ensure that a FilesFacade object is available
            if (this.filesFacade == null)
            {
                this.filesFacade = this.libManager.GetFilesFacade();
            }

            // Upload file to database
            this.filesFacade.AddIpPbxFile(filename,
                                          FileScope.Global,
                                          PreDefinedFileCategory.Skins,
                                          0,
                                          false,
                                          false,
                                          "Uploaded with CDS Client SDK FileDemo App",
                                          false,
                                          false,
                                          "en-GB",
                                          filePath,
                                          true);
        }

        //------------------------------------------------------------------------------------

        /// <summary>
        /// Returns a file list with all global and system skins
        /// </summary>
        public FileListEntrySortableCollection GetSkins()
        {
            // Ensure that a FilesFacade object is available
            if (this.filesFacade == null)
            {
                this.filesFacade = this.libManager.GetFilesFacade();
            }

            // Get skins from global scope
            FileListEntrySortableCollection globalSkins = this.filesFacade.GetFileList("%",
                FileScope.Global, PreDefinedFileCategory.Skins, 0, true).CreateSortableFilterCollection();

            // Get skins from system scope (SystemDefault)
            FileListEntrySortableCollection systemSkins = this.filesFacade.GetFileList("%",
                FileScope.SystemDefault, PreDefinedFileCategory.Skins, 0, true).CreateSortableFilterCollection();

            // Merge system and global skins into one list
            FileListEntrySortableCollection allSkins = new FileListEntrySortableCollection();

            foreach (FileListEntry entry in globalSkins)
            {
                allSkins.Add(entry);
            }

            foreach (FileListEntry entry in systemSkins)
            {
                allSkins.Add(entry);
            }

            return allSkins;
        }

        #endregion

        #region User Roles

        /// <summary>
        /// Checks if the current user has the given role
        /// </summary>
        /// <param name="role">User Role (e.g. UserSelf or Admin)</param>
        public bool HasRole(BuildInRoles role)
        {
            log.TraceInfo("HasRole", "Check for: {0}", role);

            // Ensure that a PhoneClientFacade object is available
            if (this.phoneClientFacade == null)
            {
                this.phoneClientFacade = this.libManager.GetPhoneClientFacade();
            }

            // Check if user has the role
            if (this.phoneClientFacade.HasRole(role))
            {
                // User has the role
                log.TraceVerbose("HasRole", "Success");
                return true;
            }

            // User does NOT have the role
            log.TraceVerbose("HasRole", "Failed");
            return false;
        }

        #endregion

        #region Users (Read only)

        /// <summary>
        /// Returns all users in one collection (READ ONLY)
        /// </summary>
        /// <param name="includeUserFileSize">If true the size of the user files in the database is calculated and added</param>
        public UserAdminView1EntrySortableCollection GetUserView(bool includeUserFileSize)
        {
            log.TraceInfo("GetUserView", "()");

            // Ensure that a AdminFacade object is available
            if (this.adminFacade == null)
            {
                this.adminFacade = this.libManager.GetAdminFacade();
            }
            
            try
            {
                UserDeviceSmallMap userDeviceSmallMap;
                return this.adminFacade.GetUserAdminView1(includeUserFileSize,
                        out userDeviceSmallMap).CreateSortableFilterCollection();
            }
            catch (Exception exception)
            {
                log.TraceException(exception, true);
                throw;
            }
        }

        #endregion

        #region Users (Read & Write)
       
        /// <summary>
        /// Returns a collection of all ippbx users
        /// </summary>
        public SWConfigDataClientLib.Proxies.Users.UserEntrySortableCollection GetUsers()
        {
            log.TraceInfo("GetUsers", "()");

            // If only "%" (SQL Wildcard) is used for the
            // "nameFilter" parameter all users will be returned
            return this.GetUsers("%");
        }

        //------------------------------------------------------------------------------------

        /// <summary>
        /// Returns a collection of ippbx users filtered by name
        /// </summary>
        /// <param name="nameFilter">Name of the IpPbx user your are
        /// searching for (SQL Wildcards e.g. "%" can be used here)</param>
        public SWConfigDataClientLib.Proxies.Users.UserEntrySortableCollection GetUsers(string nameFilter)
        {
            log.TraceInfo("GetUsers", "Filter: {0}", nameFilter);

            // Get UserEnum object from LibManager
            UserEnum userEnum = this.libManager.GetUserEnum();

            try
            {
                
                userEnum.ExecuteNameFilter(nameFilter, new OrderByList());
                return userEnum.PrimaryCollection.CreateSortableFilterCollection();
            }
            catch (Exception ex)
            {
                log.TraceException(ex, true);
                throw;
            }
            finally
            {
                // IMPORTANT
                // If you free resources here you will NOT be able
                // to save changes anymore. Every UserEntry has a
                // reference to his parent UserEnum to save changes.
                // If the Enum is released the reference will be broken.
                // For "read-only" purposes the AdminFacade should be
                // used because it is much faster and lightweight.

                /* Free resources
                if (userEnum != null)
                {
                    userEnum.FreeForReuse();
                }
                */
            }
        }

        //------------------------------------------------------------------------------

        /// <summary>
        /// Sets a new comment for all IpPbx user
        /// </summary>
        /// <param name="comment">New Comment</param>
        public void EditUserCommentForAllUsers(string comment)
        {
            log.TraceInfo("EditUserCommentForAllUsers", "Comment: {0}", comment);

            // Get UserEnum object from LibManager
            UserEnum userEnum = this.libManager.GetUserEnum();

            try
            {
                // Fill PrimaryCollection with all users
                userEnum.ExecuteNameFilter("%", new OrderByList());

                foreach (SWConfigDataClientLib.Proxies.Users.UserEntry userEntry in
                         userEnum.PrimaryCollection)
                {
                    userEntry.Comment = comment;
                }

                // Save all changes
                userEnum.Update();                
            }
            catch(Exception exception)
            {
                log.TraceException(exception, true);
                throw;
            }
            finally
            {
                // Free resources
                if (userEnum != null)
                {
                    userEnum.FreeForReuse();
                }
            }
        }

        //------------------------------------------------------------------------------

        /// <summary>
        /// Sets a new comment for a single IpPbx user
        /// </summary>
        /// <param name="username">Name of the user</param>
        /// <param name="comment">New comment for the user</param>
        public void EditUserCommentForOneUser(string username, string comment)
        {
            log.TraceInfo("EditUserCommentForAllUsers",
                "Username: {0} / New Comment: {1}", username, comment);

            // Get UserEnum object from LibManager
            UserEnum userEnum = this.libManager.GetUserEnum();

            try
            {
                userEnum.ExecuteNameFilter(username, new OrderByList());

                if (userEnum.PrimaryCollection.Count == 0 ||
                    userEnum.PrimaryCollection.Count > 1)
                {
                    throw new Exception("None or more than one user found. Process aborted.");
                }

                foreach (SWConfigDataClientLib.Proxies.Users.UserEntry userEntry in
                         userEnum.PrimaryCollection)
                {
                    userEntry.Comment = comment;
                    userEntry.Update(true);
                }
            }
            catch (Exception exception)
            {
                log.TraceException(exception, true);
                throw;
            }
            finally
            {
                if (userEnum != null)
                {
                    userEnum.FreeForReuse();
                }
            }
        }

        #endregion

        #region Caller List

        /// <summary>
        /// Returns a UserPhoneCallListEntrySortableCollection for an ippbx user
        /// </summary>
        /// <param name="userId">Id of the ippbx user (use 0 if you
        /// want to get the list of the current user)</param>
        public SWConfigDataClientLib.Proxies.PhoneClient.UserPhoneCallListEntrySortableCollection GetCallerList(int userId)
        {
            log.TraceInfo("GetCallerList", "({0})", userId);

            // Ensure that a PhoneClientFacade object is available
            if (this.phoneClientFacade == null)
            {
                this.phoneClientFacade = this.libManager.GetPhoneClientFacade();
            }

            try
            {
                return this.phoneClientFacade.GetUserPhoneCallList(userId, false).CreateSortableFilterCollection();
            }
            catch (Exception ex)
            {
                log.TraceException(ex, true);
                throw;
            }
        }

        //------------------------------------------------------------------------------------

        /// <summary>
        /// Clear the Caller List of an ippbx user
        /// </summary>
        /// <param name="userId">Id of the ippbx user (use 0 if
        /// you want to clear the list of the current user)</param>
        public bool ClearCallerList(int userId)
        {
            log.TraceInfo("ClearCallerList", "({0})", userId);

            // Ensure that a PhoneClientFacade object is available
            if (this.phoneClientFacade == null)
            {
                this.phoneClientFacade = this.libManager.GetPhoneClientFacade();
            }

            try
            {
                this.phoneClientFacade.DeleteUserPhoneCalls(userId);
                return true;
            }
            catch (Exception ex)
            {
                log.TraceException(ex, true);
                throw;
            }
        }

        #endregion

        #region Redial List

        /// <summary>
        /// Returns a RedialItemEntrySortableCollection for an ippbx user
        /// </summary>
        /// <param name="userId">Id of the ippbx user (use 0 if
        /// you want to get the list of the current user)</param>
        public SWConfigDataClientLib.Proxies.PhoneClient.RedialItemEntrySortableCollection GetRedialList(int userId)
        {
            log.TraceInfo("GetRedialList", "({0})", userId);

            // Ensure that a PhoneClientFacade object is available
            if (this.phoneClientFacade == null)
            {
                this.phoneClientFacade = this.libManager.GetPhoneClientFacade();
            }

            try
            {                
                return this.phoneClientFacade.GetRedialItems(userId).CreateSortableFilterCollection();
            }
            catch (Exception ex)
            {
                log.TraceException(ex, true);
                throw;
            }
        }

        //------------------------------------------------------------------------------------

        /// <summary>
        /// Clear the Redial List of an ippbx user
        /// </summary>
        /// <param name="userId">Id of the ippbx user (use 0 if
        /// you want to clear the list of the current user)</param>
        public bool ClearRedialList(int userId)
        {
            log.TraceInfo("ClearRedialList", "({0})", userId);

            // Ensure that a PhoneClientFacade object is available
            if (this.phoneClientFacade == null)
            {
                this.phoneClientFacade = this.libManager.GetPhoneClientFacade();
            }

            try
            {
                this.phoneClientFacade.DeleteRedialItems(userId);
                return true;
            }
            catch (Exception ex)
            {
                log.TraceException(ex, true);
                throw;
            }
        }

        #endregion

        #region Active Calls List

        /// <summary>
        /// Returns a list of the active calls
        /// </summary>
        public ActiveCallEntrySortableCollection GetActiveCalls()
        {
            log.TraceInfo("GetActiveCalls", "()");

            // Ensure that a AdminFacade object is available
            if (this.adminFacade == null)
            {
                this.adminFacade = this.libManager.GetAdminFacade();
            }

            try
            {
                return this.adminFacade.GetActiveCallsView().CreateSortableFilterCollection();
            }
            catch (Exception ex)
            {
                log.TraceException(ex, true);
                throw;
            }
        }

        #endregion

        #region Phonebook

        /// <summary>
        /// Returns the PBX Phonebook
        /// </summary>
        /// <returns>PBX Phonebook as PBXPhoneBookPrimaryCollection</returns>
        public PBXPhoneBookPrimaryCollection GetPbxPhonebook()
        {
            log.TraceInfo("GetPbxPhonebook", "()");

            UserPhoneBookEnum userPhoneBookEnum = this.libManager.GetUserPhoneBookEnum();

            try
            {
                return userPhoneBookEnum.GetPBXPhoneBookEx("%", true, true, true, true);
            }
            catch (Exception ex)
            {
                log.TraceException(ex, true);
                throw;
            }
            finally
            {
                // IMPORTANT
                // If you free resources here you will NOT be able
                // to save changes anymore. Every PhonebookEntry has a
                // reference to his parent UserPhoneBookEnum to save changes.
                // If the Enum is released the reference will be broken.

                /* Free resources
                if (userPhoneBookEnum != null)
                {
                    userPhoneBookEnum.FreeForReuse();
                }
                */
            }
        }

        #endregion
    }
}